﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Threading;
using System.Threading.Tasks;

public class LockFreeStack<T> {
	private Node top = null;

	// SpinWait.SpinOnce should be added in both Push and Pop.

	public void Push(T value) {
		var node = new Node();
		node.Value = value;

		do {
			node.Next = top;
		} while (Interlocked.CompareExchange(ref top, node, node.Next) != node.Next);
	}

	public T Pop() {
		Node node;
		do {
			node = top;
			if (node == null) {
				throw new InvalidOperationException("Stack is empty.");
			}
		} while (Interlocked.CompareExchange(ref top, node.Next, node) != node);

		return node.Value;
	}

	private class Node {
		public T Value;
		public Node Next;
	}

	public void DebugPrintInfo() {
		int count = 0;

		var node = top;
		while (node != null) {
			count++;
			node = node.Next;
		}

		Console.WriteLine("  Node count: {0}", count);
	}
}

public class LockFreeStackWithCache<T> {
	private Node top = null;
	private Node cacheTop = null;

	private void PushCore(ref Node stackTop, Node node) {
		do {
			node.Next = stackTop;
		} while (Interlocked.CompareExchange(ref stackTop, node, node.Next) != node.Next);
	}

	public void Push(T value) {
		var node = PopCore(ref cacheTop);
		if (node == null) {
			node = new Node();
		}

		node.Value = value;
		PushCore(ref top, node);
	}

	private Node PopCore(ref Node stackTop) {
		Node node;
		do {
			node = stackTop;
			if (node == null) {
				return null;
			}
		} while (Interlocked.CompareExchange(ref stackTop, node.Next, node) != node);

		return node;
	}

	public T Pop() {
		var node = PopCore(ref top);
		if (node == null) {
			throw new InvalidOperationException("Stack is empty.");
		}

		var value = node.Value;
		node.Value = default(T);	// Not to have a memory leak!

		PushCore(ref cacheTop, node);

		return value;
	}

	private class Node {
		public T Value;
		public Node Next;
	}

	private int DebugCountNodesIn(Node stackTop) {
		HashSet<Node> nodesSeen = new HashSet<Node>();
		
		int count = 0;

		var node = stackTop;
		while (node != null) {
			if (nodesSeen.Contains(node)) {
				Console.WriteLine("  DebugPrintInfo(): Cycle in stack detected!");
				break;
			}
			nodesSeen.Add(node);

			count++;
			node = node.Next;
		}

		return count;
	}

	public void DebugPrintInfo() {
		Console.WriteLine("  Node count: {0}", DebugCountNodesIn(top));
		Console.WriteLine("  Cached nodes: {0}", DebugCountNodesIn(cacheTop));
	}
}

class Program {
	static void Main(string[] args) {

		int upperBound = 1000; // 1000000;
		int delta = upperBound;

		Console.WriteLine("=== LockFreeStack ===");
		{
			var stack1 = new LockFreeStack<int>();

			Parallel.For(0, upperBound, item => {
				stack1.Push(item);
				stack1.Push(item + delta);
				stack1.Push(item + 2 * delta);
				stack1.Push(item + 3 * delta);
			});

			Console.WriteLine("After pushing:");
			stack1.DebugPrintInfo();

			int sum = 0;
			Parallel.For(0, upperBound, item => {
				Interlocked.Add(ref sum, stack1.Pop());
				Interlocked.Add(ref sum, stack1.Pop());
				Interlocked.Add(ref sum, stack1.Pop());
				Interlocked.Add(ref sum, stack1.Pop());
			});

			Console.WriteLine("After poping:");
			Console.WriteLine("  Sum of values: {0}", sum);
			stack1.DebugPrintInfo();
		}

		{
			var stack1 = new LockFreeStack<int>();

			int sum = 0;
			Parallel.For(0, upperBound, item => {
				stack1.Push(item);
				stack1.Push(item + delta);
				stack1.Push(item + 2 * delta);
				Interlocked.Add(ref sum, stack1.Pop());
				stack1.Push(item + 3 * delta);
			});

			Console.WriteLine();
			Console.WriteLine("After push, push, push, pop, push repeated:");
			Console.WriteLine("  Sum of values: {0}", sum);
			stack1.DebugPrintInfo();
			
			Parallel.For(0, upperBound, item => {
				Interlocked.Add(ref sum, stack1.Pop());
				Interlocked.Add(ref sum, stack1.Pop());
				Interlocked.Add(ref sum, stack1.Pop());
			});

			Console.WriteLine("After poping:");
			Console.WriteLine("  Sum of values: {0}", sum);
			stack1.DebugPrintInfo();
		}

		//
		//
		//

		Console.WriteLine();
		Console.WriteLine("=== LockFreeStack with Node cache ===");

		{
			var stack2 = new LockFreeStackWithCache<int>();

			Parallel.For(0, upperBound, item => {
				stack2.Push(item);
				stack2.Push(item + delta);
				stack2.Push(item + 2 * delta);
				stack2.Push(item + 3 * delta);
			});

			Console.WriteLine("After pushing:");
			stack2.DebugPrintInfo();

			int sum = 0;
			Parallel.For(0, upperBound, item => {
				Interlocked.Add(ref sum, stack2.Pop());
				Interlocked.Add(ref sum, stack2.Pop());
				Interlocked.Add(ref sum, stack2.Pop());
				Interlocked.Add(ref sum, stack2.Pop());
			});

			Console.WriteLine("After poping:");
			Console.WriteLine("  Sum of values: {0}", sum);
			stack2.DebugPrintInfo();
		}

		{
			var stack2 = new LockFreeStackWithCache<int>();

			int sum = 0;
			Parallel.For(0, upperBound, item => {
				stack2.Push(item);
				stack2.Push(item + delta);
				stack2.Push(item + 2 * delta);
				Interlocked.Add(ref sum, stack2.Pop());
				stack2.Push(item + 3 * delta);
			});

			Console.WriteLine();
			Console.WriteLine("After push, push, push, pop, push repeated:");
			Console.WriteLine("  Sum of values: {0}", sum);
			stack2.DebugPrintInfo();

			
			Parallel.For(0, upperBound, item => {
				Interlocked.Add(ref sum, stack2.Pop());
				Interlocked.Add(ref sum, stack2.Pop());
				Interlocked.Add(ref sum, stack2.Pop());
			});

			Console.WriteLine("After poping:");
			Console.WriteLine("  Sum of values: {0}", sum);
			stack2.DebugPrintInfo();
		}
	}
}

